package cn.com.qmth.stmms.ms.marking.service;

import java.util.Date;
import java.util.List;
import java.util.Objects;
import java.util.OptionalDouble;

import cn.com.qmth.stmms.ms.core.domain.user.MarkUser;
import cn.com.qmth.stmms.ms.core.repository.MarkUserRepo;
import cn.com.qmth.stmms.ms.marking.config.MarkingConfig;
import cn.com.qmth.stmms.ms.marking.service.arbitration.ArbitrationService;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import cn.com.qmth.stmms.ms.core.domain.Level;
import cn.com.qmth.stmms.ms.core.domain.MarkStage;
import cn.com.qmth.stmms.ms.core.domain.MarkSubject;
import cn.com.qmth.stmms.ms.core.domain.Paper;
import cn.com.qmth.stmms.ms.core.domain.task.MarkTask;
import cn.com.qmth.stmms.ms.core.repository.LevelRepo;
import cn.com.qmth.stmms.ms.core.repository.MarkSubjectRepo;
import cn.com.qmth.stmms.ms.core.repository.MarkTaskRepo;
import cn.com.qmth.stmms.ms.core.repository.PaperRepo;

/**
 * 评卷，分档仲裁及打分服务
 * Created by zhengmin on 2016/9/23.
 */
@Service
public class MarkingService {

    private static final Logger LOG = LoggerFactory.getLogger(MarkingService.class);
    @Autowired
    private MarkingConfig markingConfig;

    @Autowired
    private DetermineLevelService determineLevelService;

    @Autowired
    private MarkTaskRepo markTaskRepo;

    @Autowired
    private MarkUserRepo markUserRepo;

    @Autowired
    private PaperRepo paperRepo;

    @Autowired
    private LevelRepo levelRepo;
    
    @Autowired
    private MarkSubjectRepo markSubjectRepo;

    /**
     * 提交分档。如果每个评卷都提交了，则进行仲裁判定，如果没有触发仲裁就定档
     *
     * @param markTask
     * @param levelCode
     * @return
     */
    @Transactional
    public synchronized MarkTask levelMark(MarkTask markTask, String levelCode) {
        //TODO 回评
        //boolean isMarked = markTask.getLevel() == null;

        //如果已经由科组长定档，则不做处理
        Paper paper = markTask.getPaper();
        if (paper.isMarkByLeader()) {
            return markTask;
        }
        /**
        阀值判断

        //控制分档比例
        List<Level> levels = levelRepo.findByWorkId(markTask.getWorkId());
        Level targetLevel = null;
        for (Level l : levels) {
            if (l.getCode().equals(levelCode)) {
                targetLevel = l;
                break;
            }
        }
        int pt = targetLevel.getPt();
        int count = markTaskRepo.countByWorkIdAndSubjectAndMarkerIdAndStageAndResult(markTask.getWorkId(), markTask.getSubject(), markTask.getMarkerId(), markTask.getStage(), levelCode);
        int total = markTaskRepo.countByWorkIdAndSubjectAndMarkerIdAndStage(markTask.getWorkId(), markTask.getSubject(), markTask.getMarkerId(), markTask.getStage());

        if ((count + 1.0D) / total * 1.0D > pt / 100.D) {
            throw new RuntimeException(levelCode + "档已满，占比总阀值已达" + pt + "%。");
        }
        //判断考点阀值是否已满
        int kdpt = targetLevel.getKdpt();
        int kdcount = markTaskRepo.countByWorkIdAndSubjectAndMarkerIdAndStageAndQuestionIdAndResult(markTask.getWorkId(), markTask.getSubject(), markTask.getMarkerId(), markTask.getStage(), markTask.getQuestionId(), levelCode);
        int kdtotal = markTaskRepo.countByWorkIdAndSubjectAndMarkerIdAndStageAndQuestionId(markTask.getWorkId(), markTask.getSubject(), markTask.getMarkerId(), markTask.getStage(), markTask.getQuestionId());
        if ((kdcount + 1.0D) / kdtotal * 1.0D > kdpt / 100.D) {
            throw new RuntimeException(levelCode + "档已满，考点阀值已达" + kdpt + "%。");
        }
         */

        MarkUser maker = markUserRepo.findOne(markTask.getMarkerId());
        Level level = levelRepo.findByWorkIdAndCode(markTask.getWorkId(), levelCode);

        markTask.setResult(levelCode);
        markTask.setLevelValue(level.getLevelValue());
        markTask.setRejected(false);
        markTask.setUpdatedOn(new Date());
        markTask.setMarkerName(maker.getName());

        markTaskRepo.save(markTask);

        //判断该任务的试卷是否所有评卷员都评完，要进行仲裁判定或定档
        List<MarkTask> markTasks = markTaskRepo.findByPaperIdAndStage(paper.getId(), markTask.getStage());
        long leftCount = markTasks.stream().filter(i -> i.getResult() == null).count();
        if (leftCount == 0) {
            MarkTask[] sources = new MarkTask[markTasks.size()];
            int[] values = markTasks.stream().mapToInt(MarkTask::getLevelValue).toArray();
            boolean result = ArbitrationService.arbitrate(markingConfig.getDeviation(), values);
            if (result) {
                //仲裁
                paper.arbitrate();
            } else {
                //定档
                DeterResult determine = determineLevelService.determine(markingConfig.isMajority(),markTasks.toArray(sources));
                paper.determineLevel(determine.getResult());
                if (determine.getDeterType() == DetermineLevelService.DeterType.WEIGHT) {
                    paper.setMarkedLogic(true);
                } else {
                    paper.setMarkedLogic(false);
                }
            }
            paperRepo.save(paper);
        }
        return markTask;
    }

    /**
     * 直接给试卷分档
     *
     * @param paper
     * @param level
     */
    @Transactional
    public Paper levelMarkPaper(Paper paper, String level,boolean isSample) {
        //2019.10.12 湖北美术统考去掉标准卷设置数量
//    	if(isSample){
//    		MarkSubject markSubject = markSubjectRepo.findOne(paper.getWorkId() + "-" + paper.getSubject().toString());
//    		long sampleCount = paperRepo.countByWorkIdAndSubjectAndIsSampleIsTrue(paper.getWorkId(), paper.getSubject());
//    		if(sampleCount >= markSubject.getSampleCount()){
//    			throw new RuntimeException("标准卷上限已达到");
//    		}
//    	}
    	paper.determineLevel(level);
    	paper.setMarkByLeader(true);
    	paper.setSample(isSample);
    	paperRepo.save(paper);
    	List<MarkTask> markTasks = markTaskRepo.findByPaperIdAndStage(paper.getId(), MarkStage.LEVEL);
    	for (MarkTask markTask : markTasks) {
    		markTask.setResult(level);
    	}
    	markTaskRepo.save(markTasks);
        return paper;
    }

    /**
     * 打回重评
     * 更新试卷状态
     * 更新评卷任务状态为打回
     * 如果评卷任务的档位和建议档位一致，不打回该任务
     *  @param paper
     * @param redoLevel 建议重评档位
     * @param ranges
     */
    @Transactional
    public Paper reject(Paper paper, String redoLevel, String ranges) {
        paper.reject(redoLevel);
        //打回后撤销标准卷
        paper.setSample(false);
        List<MarkTask> markTasks = markTaskRepo.findByPaperIdAndStage(paper.getId(), MarkStage.LEVEL);
        for (MarkTask markTask : markTasks) {
            /*if (markTask.getResult() != null && !markTask.getResult().toString().equals(redoLevel)) {
                markTask.setRejected(true);
                markTask.setOriginLevel(markTask.getResult());
                markTask.setResult(null);
            }*/
            String[] range = ranges.split(",");
            if(range.length > 0){
                for (String r : range) {
                    if(Objects.equals(String.valueOf(markTask.getMarkerId()), r)){
                        markTask.setRejected(true);
                        markTask.setOriginLevel(markTask.getResult());
                        markTask.setResult(null);
                    }
                }
            }
        }
        markTaskRepo.save(markTasks);
        paperRepo.save(paper);
        return paper;
    }

    /**
     * 打分。取出组内所有评分，如果都给了分，取平均分更新试卷成绩
     *
     * @param markTask
     * @param score
     */
    @Transactional
    public MarkTask scoring(MarkTask markTask, int score) {
        if (markTask.getStage() != MarkStage.SCORE) {
            throw new RuntimeException("该任务不是打分任务");
        }
        markTask.setResult(String.valueOf(score));
        markTask.setUpdatedOn(new Date());
        MarkUser maker = markUserRepo.findOne(markTask.getMarkerId());
        markTask.setMarkerName(maker.getName());
        markTaskRepo.save(markTask);
        Paper paper = markTask.getPaper();
        List<MarkTask> markTasks = markTaskRepo.findByPaperIdAndStage(paper.getId(), markTask.getStage());
        long leftCount = markTasks.stream().filter(i -> i.getResult() == null).count();
        if (leftCount == 0) {
            OptionalDouble finalScore = markTasks.stream().map(m -> m.getResult())
                    .mapToInt(Integer::valueOf).average();
            //湖北省，平均分非零进一
            double fs = markingConfig.isRoundUp() ? Math.ceil(finalScore.orElse(0)) : Math.round(finalScore.orElse(0));
            paper.setScore(fs);
            paperRepo.save(paper);
        }
        return markTask;
    }
    /**
     * 跳过该任务，最后处理
     * @param markTaskId
     */
    @Transactional
	public MarkTask skip(Long markTaskId) {
		try {
			MarkTask m = markTaskRepo.findOne(markTaskId);
			Integer randomSeq = markTaskRepo.findRandomSeqByWorkIdAndSubjectAndMarkerIdAndStageAndQuestionId(m.getWorkId(), m.getSubject(),m.getMarkerId(), m.getStage(), m.getQuestionId());
			if(randomSeq == null){
				randomSeq = 0;
			}
			m.setRandomSeq(randomSeq+1);
			markTaskRepo.save(m);
			return m;
		} catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("该任务可能被占用，跳过失败");
		}
		
	}
}
